--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

FauxMazzleUI = {}

local FMFrame, FMDisabled

local find, format, lower, sub, upper = string.find, string.format, string.lower, string.sub, string.upper
local ceil, floor = math.ceil, math.floor

NS.UIW, NS.UIH = floor(UIParent:GetWidth()), floor(UIParent:GetHeight())

NS.DebugPrint = FIZZLEDEBUG and FIZZLEDEBUG(AddonName) or function() end

local L = NS.TEXTS

local ParentFrameName, GlobalName = AddonName.."Parent", upper(AddonName)
local SetupType = {
	Character		= 0x1,
	CharacterDefaults	= 0x2,
	Initialise		= 0x4,
	InitialiseAddons	= 0x8,
	NewSettings		= 0x10,
	OpenDialog		= 0x20,
	Reset			= 0x40,
	HighlightReload		= 0x80,
}

NS.PortaitTypes = { 
	["3D"]=0,
	["2D"]=1,
	["Text"]=2,
}

-- Map versions that may not exist in the client yet. 
-- MAINLINE should always exist, old clients may not have newer project IDs.
NS.WOW_Project = setmetatable({}, NS.readmeta)
NS.WOW_Project._metaError = "Missing WOW_PROJECT option in NS.WOW_Project table:" -- "'_metaError' is printed out during /fm opts
NS.WOW_Project.Classic = WOW_PROJECT_CLASSIC or 2
NS.WOW_Project.Burning_Crusade_Classic = WOW_PROJECT_BURNING_CRUSADE_CLASSIC or 5
NS.WOW_Project.WrathClassic = WOW_PROJECT_WRATH_CLASSIC or 11

function NS:IsValidVersion(lower, higher, include_retail)
	if not lower or type(lower) ~= "number" then -- valid for all versions
		NS.DebugPrint("IsValidVersion Error arg1:", lower)
		return false
	end
	if type(higher) == "boolean" then
		include_retail = higher
		higher = nil
	end
	if WOW_PROJECT_ID == WOW_PROJECT_MAINLINE then
		if (higher and higher == WOW_PROJECT_MAINLINE) or lower == WOW_PROJECT_MAINLINE or include_retail then
			return true
		end
	end
	if higher then
		if WOW_PROJECT_ID >= lower and WOW_PROJECT_ID <= higher then -- All version above lower + MainLine
			return true
		end
	else
		if lower == WOW_PROJECT_ID then
			return true
		end
	end
	return false
end

---------------------------
--    Media Locations    --
---------------------------
local Path = "Interface\\Addons\\"..AddonName.."\\"
NS.Font = "Fonts\\FRIZQT__.TTF" -- Default font for now
NS.SharedMediaFont = "Arial Narrow"
NS.FontPath = Path.."Fonts\\" -- path to non build-in fonts
NS.FontSize = 12
NS.TexturePath = Path.."Textures\\"

if NS:IsValidVersion(NS.WOW_Project.Classic) then
	NS.FontSize = 8
end
if NS:IsValidVersion(NS.WOW_Project.Burning_Crusade_Classic) then
	NS.FontSize = 9
end
--------------------------
--   Shared Media Lib.  --
--------------------------
local  media = LibStub:GetLibrary("LibSharedMedia-3.0")
media:Register("statusbar", "BantoBarReverse", NS.TexturePath.."BantoBarReverse")

----------------------------------
--  General Functions and Meta  --
----------------------------------
if not GetAddOnMetadata then -- 10.1.0 change (GetAddOnMetadata wasn't added to Deprecated on the PTR)
	local GetAddOnMetadata = C_AddOns.GetAddOnMetadata
end

local function ShowGameMeta()
	print("|cff00ff00"..AddonName.."|r |cffffff00Version:|r", GetAddOnMetadata(AddonName, "Version"))
	print("WoW Version:", WOW_PROJECT_ID)
end

----------------------------------
--   Functions used by Options  --
----------------------------------
local Options = setmetatable({}, NS.readmeta)
Options._metaError = "Missing NS:AddOption func." -- "'_metaError' is printed out during /fm opts
function NS:AddOption(name, value)
	name = upper(name)
	Options[name] = value
	return true
end

function NS:GetOption(name)
	name = upper(name)
	if not Options[name] then
		L.print("NS:GetOption", L.ErrorNoOption, name)
		return
	end
	return Options[name]
end
NS:AddOption("FauxMazzle", AddonName)
NS:AddOption("Font", NS.Font)
NS:AddOption("SharedMediaFont", NS.SharedMediaFont)
NS:AddOption("FontSize", NS.FontSize)
NS:AddOption("FontPath", NS.FontPath)
NS:AddOption("PortaitTypes", NS.PortaitTypes)
NS:AddOption("SetupType", SetupType)
NS:AddOption("TexturePath", NS.TexturePath)
NS:AddOption("Texts", L)
NS:AddOption("AddOption", NS.AddOption)
NS:AddOption("IsValidVersion", NS.IsValidVersion)
NS:AddOption("WowProjects", NS.WOW_Project)

local function ShowOptions()
	for k,_ in pairs(Options) do
		print(k)
	end
end


function NS:GetScale(multiplier)
	local RetVal = 1
	if NS:IsValidVersion(NS.WOW_Project.Classic, NS.WOW_Project.Burning_Crusade_Classic) then
--	if NS:IsValidVersion(NS.WOW_Project.Classic) then
		RetVal = 0.71
	end
	if multiplier then
		RetVal = RetVal * multiplier
	end
	return RetVal
end

NS:AddOption("GetScale", NS.GetScale)

----------------------------------
--      Accessory Information	--
----------------------------------

local Accessories, DefaultAccessories

local function InitAccessories()
	Accessories = {
		skin={},
		meter={
			damage={},
			threat={},
		},
	}
	DefaultAccessories = {
		skin="FauxMazzle_Skin_Default",
		meter={
			damage="FauxMazzle_Recount",
			threat="FauxMazzle_Omen",
		},
	}
end

local function LoadAccessoryList()
	if Accessories then return end
	InitAccessories(Accessories)
	InitAccessories(DefaultAccessories, true)
	local AddonCount = GetNumAddOns()
	for i=1, AddonCount do
		local Folder = GetAddOnInfo(i)
		local Enabled = GetAddOnEnableState(UnitName("player"), i)
		local Group = GetAddOnMetadata(i, "X-FauxMazzleGroup")
		Group = Group and lower(Group)
		if Enabled == 2 and Accessories[Group] then
			local Found
			local SubGroup = GetAddOnMetadata(i, "X-FauxMazzleSubGroup")
			SubGroup = SubGroup and lower(SubGroup)
			if SubGroup then
				if Accessories[Group][SubGroup] then
					Accessories[Group][SubGroup][Folder] = {}
					Found = Accessories[Group][SubGroup][Folder]
				end
			else
				Accessories[Group][Folder] = {}
				Found = Accessories[Group][Folder]
			end
			if Found then
				Found.title = GetAddOnMetadata(i, "X-FauxMazzleTitle") or Folder
				Found.author = GetAddOnMetadata(i, "X-FauxMazzleAuthor") or GetAddOnMetadata(i, "Author") or "Unknown"
				Found.author = " |cffffffc0["..Found.author.."]|r"
			end
		end
	end
	
end

function NS.GetAccessoryInfo(folder, group, subgroup)
	LoadAccessoryList()
	if not group then
		L.print(L.GetText(L.BadParameter, "group", "GetAccessoryInfo", "nil"))
		return nil
	end
	group = group and lower(group)
	subgroup = subgroup and lower(subgroup)
	local Accessory = Accessories[group]
	if not Accessory then
		L.print(L.GetText(L.BadParameter, "group", "GetAccessoryInfo", "nil"))
		return nil
	end	
	if subgroup then
		Accessory = Accessory[subgroup]
	end
	if not Accessory then
		L.print(L.GetText(L.BadParameter, "subgroup", "GetAccessoryInfo", subgroup and subgroup or "nil"))
		return nil
	end
	Accessory = Accessory[folder]
	if not Accessory then
		L.print("NS.GetAccessoryInfo", L.GetText(L.AccessoryNotFound, L.Addon), folder)
		return nil
	end
	return Accessory
end
NS:AddOption("GetAccessoryInfo", NS.GetAccessoryInfo)

function NS.GetAccessoryList(group, subgroup)
	LoadAccessoryList()
	group = group and lower(group)
	subgroup = subgroup and lower(subgroup)
	if not group then
		return Accessories
	end
	if Accessories[group] then
		if subgroup then
			if Accessories[group][subgroup] then
				return Accessories[group][subgroup]
			end
		else
			return Accessories[group]
		end
	end
	return
end
NS:AddOption("GetAccessoryList", NS.GetAccessoryList)

function NS.GetAccessorySettings(folder, group, subgroup)
	local AccessoryInfo = NS.LoadAccessory(folder, group, subgroup)
	if AccessoryInfo then
		AccessoryInfo = AccessoryInfo.settings
	end
	return AccessoryInfo
end
NS:AddOption("GetAccessorySettings", NS.GetAccessorySettings)

local function DefaultAccessoryError(group, subgroup)
	if subgroup then
		L.print("DefaultAccessoryError", L.GetText(L.AccessoryNotFound, L.Group.."["..L.Subgroup.."]"), group.."["..subgroup.."]")
	else
		L.print("DefaultAccessoryError", L.GetText(L.AccessoryNotFound, L.Group), group)
	end
end

function NS.GetDefaultAccessory(group, subgroup)
	LoadAccessoryList()
	if not group then
		L.print(L.GetText(L.BadParameter, "group", "GetDefaultAccessory", group and group or "nil"))
		return
	end
	group = lower(group)
	subgroup = subgroup and lower(subgroup)
	local Accessory = DefaultAccessories[group]
	if not Accessory then
		L.print(L.GetText(L.BadParameter, "group", "GetDefaultAccessory", group))
		return
	end
	if subgroup then
		Accessory = Accessory[subgroup]
		if not Accessory then
			L.print(L.GetText(L.BadParameter, "subgroup", "GetDefaultAccessory", subgroup))
			return
		end
	end
	local SeekList = Accessories[group]
	if not SeekList then
		DefaultAccessoryError(group)
		return
	end
	if subgroup then
		SeekList = Accessories[group][subgroup]
		if not SeekList then
			DefaultAccessoryError(group, subgroup)
			return
		end
	end
	if not SeekList[Accessory] then -- get the first available to use as default ie, they don't like mine.
		local key, value = next(SeekList)
		if not key then
			return
		end
		Accessory = key
	end
	return Accessory
end
NS:AddOption("GetDefaultAccessory", NS.GetDefaultAccessory)

function NS.GetDefaultAccessoryList(group) -- Only really usefull for keys with multiple sub-goups
	if not group then
		L.print(L.GetText(L.BadParameter, "group", "GetDefaultAccessoryList", "nil"))
		return
	end
	group = lower(group)
	local Accessory = DefaultAccessories[group]
	if not Accessory then 
		L.print(L.GetText(L.BadParameter, "group", "GetDefaultAccessoryList", group))
	end
	return Accessory
end
NS:AddOption("GetDefaultAccessoryList", NS.GetDefaultAccessoryList)

function NS.LoadAccessory(folder, group, subgroup)
	LoadAccessoryList()
	local Folder = folder
	local AccessoryInfo = NS.GetAccessoryInfo(Folder, group, subgroup)
	if AccessoryInfo and AccessoryInfo.loaded then
		return AccessoryInfo
	end
	local loaded, reason = LoadAddOn(Folder)
	if not loaded or not FauxMazzleUI[Folder] then
		if reason then
			L.print("NS.LoadAccessory", L.AccessoryNotLoaded, Folder, _G["ADDON_"..reason])
		end
		Folder = NS.GetDefaultAccessory(group, subgroup)
		if not Folder then
			L.print("NS.LoadAccessory", L.AccessoryNoDefault, group, subgroup)
			return
		end
		loaded, reason = LoadAddOn(Folder)
		if not loaded or not FauxMazzleUI[Folder] then
			if reason then
				L.print("NS.LoadAccessory", L.AccessoryNotLoaded, Folder, _G["ADDON_"..reason])
			end
			return
		end
		AccessoryInfo = NS.GetAccessoryInfo(Folder, group, subgroup)
	end
	AccessoryInfo.loaded = true
	AccessoryInfo.settings = AccessoryInfo.settings or FauxMazzleUI[Folder]
	table.wipe(FauxMazzleUI)
	return AccessoryInfo
end
NS:AddOption("LoadAccessory", NS.LoadAccessory)

----------------------------------
--      Utility Functions       --
----------------------------------
-- General
function NS:GetSettings(branch, profile, surpress)
	if type(profile) == "boolean" then
		surpress = profile
		profile = nil
	end
	local Settings = _G[GlobalName]
	if not branch then
		if profile then
			profile = upper(sub(profile, 1, 1)) .. lower(sub(profile, 2))
			Settings = Settings[profile]
		end
		return Settings
	end
	branch = upper(branch)
	local RetVal
	if branch == "PROFILE" then
		if profile then
			return Settings.Profiles[profile]-- not current character
		else
			return Settings.Profiles[NS:GetOption("profile")] -- current character all settings
		end
	elseif branch == "PROFILES" then
		return Settings.Profiles
	elseif Settings[branch] then -- global branch settings
		return Settings[branch]
	else
		RetVal = NS:GetSettings("Profile", profile)[branch] -- current character branch settings
	end
	if RetVal then
		return RetVal
	end
	if not surpress then
		L.print("NS:GetSettings()", "Not Implimented!", branch, profile)
	end
end
NS:AddOption("GetSettings", NS.GetSettings)

function NS:GetCharacterSettings(branch, leaf)
	local settings = _G[GlobalName.."Char"]
	if not branch then 
		return settings
	end
	branch = upper(branch)
	if not settings[branch] then
		settings[branch] = {}
	end
	settings = settings[branch]
	if leaf then
		if not settings[leaf] then
			settings[leaf] = {}
		end
		settings = settings[leaf]
	end
	return settings
end
NS:AddOption("GetCharacterSettings", NS.GetCharacterSettings)

-- Frames
function NS:GetFrameStack(tbl)
	return tbl.Strata, tbl.Level, tbl.Layer, tbl.SubLayer
end

function NS:SetFrameStack(frame, tbl, strata, level, layer, sublayer)
	local Strata = tbl.strata or strata
	if Strata then
		frame:SetFrameStrata(Strata)
	end
	local Level = tbl.level or level
	if Level then
		frame:SetFrameLevel(Level)
	end
	local Layer = tbl.layer or layer
	local SubLayer = tbl.sublayer or sublayer
	if Layer then
		frame:SetDrawLayer(Layer, SubLayer)
	end
end

function NS:GetFrameParent(tbl)
	if not tbl.parent then
		return FMFrame
--		return UIParent
	elseif tbl.parent.stack then
		return NS:GetFrame(tbl.parent.stack, tbl.parent.frame)
	elseif tbl.parent.frame then
		return tbl.parent.frame
	else
		NS.DebugPrint("Framework.lua:", "NS:GetFrameParent() error!")
	end
end

local function AttachFrame(frame, points)
	local relative
	if not points.relative then
		relative = frame:GetParent()
	elseif points.stack then
		relative = NS:GetFrame(points.stack, points.relative)
	elseif type(points.relative) == "string" then
		relative = _G[points.relative]
	else
		relative = points.relative
	end
	frame:SetPoint(points.point, relative, points.topoint or points.point, points.x or 0, points.y or 0)
end

function NS:SetPoints(frame, tbl)
	if type(frame) == "string" then
		frame = _G[frame]
	end
	frame:ClearAllPoints()
	for i=1, #tbl do
		AttachFrame(frame, tbl[i])
	end
	if frame.Text and tbl.text then
		AttachFrame(frame.Text, tbl.text)
	end
end

local SavePoints = {}

function NS:SavePoints(frame, restore)
	if not SavePoints[frame] then
		SavePoints[frame] = { points={} }
		for i=1, 4 do
			local p, r, t, x, y = frame:GetPoint(i)
			if p and p ~= "" then
				SavePoints[frame].points[i] = { p, r, t, x, y}
			end
		end
	end
	if restore then
		for k, v in pairs(SavePoints[frame].points) do
			NS.PointFrame.SetPoint(frame, unpack(v))
		end
	end
end

local function SetDissapeard(self)
	if self.FMHooked then return end
	FMFrame.ClearAllPoints(self)
	FMFrame.SetPoint(self, "TOP", FauxMazzleRemoved, "BOTTOM")
end
function NS:DisappearFrame(frame, reset)
	if not FauxMazzleRemoved then
		CreateFrame("Frame", "FauxMazzleRemoved")
		FauxMazzleRemoved:SetSize(5, 5)
		FauxMazzleRemoved:SetPoint("TOP", UIParent, "BOTTOM", 0, -50)
		FauxMazzleRemoved:Hide()
	end
--[[
	if not frame.OldParent then
		frame.OldParent = frame:GetParent()
--		NS:Replacepoints(frame)
		frame:HookScript("OnShow", function(self)
			if self.FmDisapeared then
				self:Hide()
			end
		end)
	end
]]--
	if frame.FmDisapeared and not reset then return end
	SetDissapeard(frame)
--[[
	if not reset then
		frame:SetParent(FauxMazzleRemoved)
		frame:OldClearAllPoints()
		frame:OldSetPoint("TOP", FauxMazzleRemoved, "BOTTOM")
	else
		frame:SetParent(frame.OldParent)
		frame:OldClearAllPoints()
		NS:SavePoints(frame, true)
	end
]]--
	frame.FmDisapeared = not reset
end

function NS:TimeLeft(time, endcooldown, hideprefix, notify)
	local timeleft = endcooldown and endcooldown - time or time
	if timeleft < 60 then
		timeleft = ceil(timeleft)
		return format("%d%s", timeleft, hideprefix and "" or L.AbrevSecond ), notify and timeleft <= notify
	else 
		timeleft = timeleft/60
		if timeleft < 60 then
			timeleft = ceil(timeleft)
			return format("%d%s", timeleft, L.AbrevMinute)
		else
			timeleft = timeleft/60
			if timeleft < 24 then
				timeleft = ceil(timeleft)
				return format("%d%s", timeleft, L.AbrevHour)
			end
		end
	end
	return format("24+%s",L.AbrevHour, L.AbrevHour)
end

--------------------------
----- Slash Command  -----
--------------------------
local function ToggleOptions(setuptype, transition)
	if not IsAddOnLoaded("FauxMazzleConfig") then
		local loaded, reason = LoadAddOn("FauxMazzleConfig")
		if not loaded then
			L.print("ToggleOptions", L.GetText(L.ErrorLoadOptions, _G["ADDON_"..reason]))
			return
		end
	end
	FauxMazzleConfig(NS.GetOption, setuptype, transition)
end

_G["SLASH_"..GlobalName.."1"] = "/fauxmazzle"
_G["SLASH_"..GlobalName.."2"] = "/fm"
SlashCmdList[GlobalName] = function(msg)
	if upper(msg) == "OPT" and FIZZLEDEBUG then
		ShowGameMeta()
		ShowOptions()
		return
	elseif upper(msg) == "RESET" then
		local Profile = NS:GetSettings("Profiles")
		Profile[NS:GetOption("profile")] = nil
		Profile = NS:GetCharacterSettings()
		Profile = nil
		ReloadUI()
		return
	elseif upper(msg) == "PORT" and FIZZLEDEBUG then
		NS:SendPortraitToStudio()
		return
	end
	ToggleOptions(SetupType.OpenDialog)
end

--------------------------
-----   ON Update    -----
--------------------------

local UpdateSpeeds = {
	[1]=0.0333, --1/30,
	[2]=0.0666, --1/15,
	[3]=0.125, --1/8,
	[4]=0.12, --1/5,
	[5]=1,
	[6]=3,
	[7]=10,
	[8]=60,
}

local OnUpdateFuncs = {}
local Elapsed = {}

local function OnUpdate(self, elapsed)
	for elapseId, elapseTime in pairs(Elapsed) do
		elapseTime = elapseTime - elapsed
		if elapseTime > 0 then
			Elapsed[elapseId] = elapseTime
		else
			Elapsed[elapseId] = UpdateSpeeds[elapseId]
			if OnUpdateFuncs[elapseId] then
				for k, v in pairs(OnUpdateFuncs[elapseId]) do
					if not self:IsShown() then return end
					if v.runhidden or k:IsShown() then
						v.callback(k)
					end
				end
			end
		end
	end
end

function NS:RegisterForUpdates(frame, frequency, callback, runhidden)
	if not Elapsed[frequency] then
		Elapsed[frequency] = UpdateSpeeds[frequency]
	end
	if not OnUpdateFuncs[frequency] then
		OnUpdateFuncs[frequency] = {}
	end
	OnUpdateFuncs[frequency][frame] = { callback=callback, runhidden=runhidden }
	callback(frame)
end

function NS:UnregisterForUpdates(frame, frequency)
	if OnUpdateFuncs[frequency] and OnUpdateFuncs[frequency][frame] then
		OnUpdateFuncs[frequency][frame] = nil
		if not next(OnUpdateFuncs[frequency]) then
			OnUpdateFuncs[frequency] = nil
			Elapsed[frequency] = nil
		end
	end
end

--------------------------
-----   Frame Store  -----
--------------------------
local Frames = {}
function NS:AddFrame(branch, name, frame, override)
	branch = upper(branch)
	name = upper(name)
	if not Frames[branch] then
		Frames[branch] = {}
	end
	if Frames[branch][name] and not override then
		L.print("NS:AddFrame", "Branch[name] exists!", branch, name)
		return
	end
	Frames[branch][name] = frame
	return frame
end
NS:AddOption("AddFrame", NS.AddFrame)

function NS:GetFrame(branch, name, suppress)
	branch = upper(branch)
	if not Frames[branch] then
		if not suppress then
			L.print("NS:GetFrame", "Branch does not exist!", branch, "for name:", name)
		end
		return
	end
	if not name then
		return Frames[branch]
	end
	name = upper(name)
	if not Frames[branch][name] then
		if not suppress then
			L.print("NS:GetFrame", "Name does not exist!", name, "for branch:", branch)
		end
		return
	end
	return Frames[branch][name]
end
NS:AddOption("GetFrame", NS.GetFrame)

function NS:RemoveFrame(branch, name)
	if not (branch or name) then
		L.print("NS:RemoveFrame", "Parameter missing!", branch, name)
	end
	branch = upper(branch)
	name = upper(name)
	if not Frames[branch] then
		L.print("NS:RemoveFrame", "Branch does not exist!", branch)
		return
	end
	if not Frames[branch][name] then
		L.print("NS:RemoveFrame", "Name does not exist!", name)
		return
	end
	Frames[branch][name] = nil
end
NS:AddOption("RemoveFrame", NS.RemoveFrame)

--------------------------
-----  Event Handler -----
--------------------------
local Events = {}

local EventFrame = CreateFrame("Frame")

local function FauxMazzleOnEvent(self, event, ...)
	for frame, framelist in pairs(Events[event]) do
		for func,_ in pairs(framelist) do
			func(frame, event, ...)
		end
	end
end

function NS:PostEvent(event, ...)
	FauxMazzleOnEvent(EventFrame, event, ...)
end

function NS:RegisterEvent(event, frame, func)
	if FMDisabled then return end
	if not Events[event] then
		Events[event] = {}
		EventFrame:RegisterEvent(event)
	end
	if not Events[event][frame] then
		Events[event][frame] = {}
	end
	if not Events[event][frame][func] then
		Events[event][frame][func] = true
	end
end

function NS:UnregisterEvent(event, frame, func)
	if not Events[event] or FMDisabled then return end
	if Events[event] and Events[event][frame] and Events[event][frame][func] then
		Events[event][frame][func] = nil
	end
	local k, v = next(Events[event][frame])
	if v then
		return
	end
	Events[event][frame] = nil
	k, v = next(Events[event])
	if v then
		return
	end
	EventFrame:UnregisterEvent(event)
	Events[event] = nil
end

function NS:UnregisterAllEvents(frame)
	for event, eventlist in pairs(Events) do
		for frame, framelist in pairs(Events[event]) do
			if not frame or frame == frame then
				for func,_ in pairs(framelist) do
					NS:UnregisterEvent(event, frame, func)
				end
			end
		end
	end
end

--------------------------
-----  Module Inits. -----
--------------------------
local InitEventList = {
	ADDON_LOADED = {},
	PLAYER_ENTERING_WORLD = {},
	PLAYER_LOGIN = {},
	PLAYER_LOGOUT = {},
	UPDATE_CHAT_WINDOWS = {},
}

function NS:RegisterForInit(event, name, priority, func, lower, higher, include_retail)
	if lower then
		if not NS:IsValidVersion(lower, higher, include_retail) then
			return
		end
	end
	if not InitEventList[event] then
		InitEventList[event] = {}
		FMFrame:RegisterEvent(event)
	end
--	local args = { ... }
--	if not name or not priority or not func or not args[1] then
	if not name or not priority or not func then
		NS.DebugPrint("NS:RegisterForInit Parameter missing", name, priority, func, lower, higher, include_retail)
		return
	end
	tinsert(InitEventList[event], { name=name, priority=priority, func=func } )
end
NS:AddOption("RegisterForInit", NS.RegisterForInit)

--[[
-- Not a Replacement for DiscordLib_RegisterInitFunc(). Needs replacing
-- DiscordLib_RegisterInitFunc waits for VARIABLES_LOADED AND PLAYER_ENTERING_WORLD before adding delay and firing
local Variables_Loaded, Player_Entered_World
local function CheckInit(event
function NS:RegisterForDelayedInit(func, ...)
	if not ValidVersion(...) then
		return 
	end
	tinsert(DL_INIT_FUNCTIONS, func)
end
NS:AddOption("RegisterForDelayedInit", NS.RegisterForDelayedInit)
]]--

local function SortTable(t, key, descending)
	sort(t, function(a, b)
		if descending then
			return a[key] > b[key]
		end
		return a[key] < b[key]
	end)
end
NS:AddOption("SortTable", SortTable)

local function RunInit(self, event, ...)
	local t = InitEventList[event]
	SortTable(t, "priority")
	for k, v in ipairs(t) do
		v.func(self, ...)-- Original
	end
	InitEventList[event] = nil -- Only fire once for init
end

function NS:GetFontMetrics(fonttype, font, size, text)
	local f = FMFrame
	if fonttype == 0 then
		f.Text:SetFontObject(font)
	else
		f.Text:SetFont(font, size, "")
	end
	if not text or text == "" then
		f.Text:SetText("WW")
	else
		f.Text:SetText(text)
	end
	return ceil(f.Text:GetWidth()), ceil(f.Text:GetHeight())
end
NS:AddOption("GetFontMetrics", NS.GetFontMetrics)

-------------------------------------- Get Started -----------------------------------
FMFrame = CreateFrame("Frame", "FauxMazzleFrame", UIParent)
NS.PointFrame = FMFrame
FMFrame:SetSize(5, 5)
FMFrame.Text = FMFrame:CreateFontString()
FMFrame.Text:Hide()
FMFrame.Text:SetPoint("LEFT")
FMFrame.Text:SetFont(NS.Font, NS.FontSize, "")

FMFrame:SetScript("OnEvent", function(self, event, ...)
	self:UnregisterEvent(event)
	if FMDisabled then return end
	RunInit(self, event, ...)
	if event == "PLAYER_ENTERING_WORLD" then
 		self:SetScript("OnUpdate", OnUpdate)
 		NS:PostEvent("PLAYER_LOGIN") -- Init for registered events that are character specific
 		NS:PostEvent("PLAYER_ENTERING_WORLD") -- Becsause we passed over this event first time in
 		EventFrame:SetScript("OnEvent", FauxMazzleOnEvent) -- The world is ready... we hope
	end
end)

FMFrame:SetScript("OnHide", function(self)
	if not UIParent:IsShown() then
	    NS:SetViewport(true)
	end
end)

FMFrame:SetScript("OnShow", function(self)
	if UIParent:IsShown() then
	    NS:SetViewport()
	end
end)


for k,_ in pairs(InitEventList) do
	FMFrame:RegisterEvent(k)
end

-------------------------------------- Environments -----------------------------------
local NewSettings = { -- update number[id], percharacter update true/false
	[1]=false,
	[2]=false,
	[3]=false,
	[4]=true,
	[5]=true,
	[6]=true,
	[7]=true,
	[8]=true,
	[9]=false,
	[10]=false,
	[11]=false,
}
NS:AddOption("NewSettings", NewSettings)

--------------------------------------     World    -----------------------------------
local function InitialiseEnvironmentWorld(self, ...)
	local OpenConfig
	NS:AddOption("Profile", UnitName("player").."-"..GetRealmName())
	NS.AceProfile = UnitName("player").." - "..GetRealmName()
	NS:AddOption("GlobalName", GlobalName)

	local Settings = NS:GetSettings()
	if not Settings or FAUXMAZZLE_Options then -- new install so call options to initialise everything including FM1 upgrade and open the GUI
		ToggleOptions(SetupType.Initialise)
		OpenConfig = true
	end
--------------------------------------   Character    -----------------------------------
	Settings = NS:GetSettings("Profile")
	if not Settings then -- new character also opens the GUI (if not new install)
		ToggleOptions(SetupType.Character)
		OpenConfig = true
	end
	if NS:GetSettings("Disabled", true) then -- FauxMazzle disabled for this character true = Supress FizzleDebug error message
		FMDisabled = true
		return 
	end
------------------------------------   New Settings    ----------------------------------
	local NewSettings = NS:GetOption("NewSettings")
	local LatestVersion = #NewSettings
	local LastGlobalUpdate = NS:GetSettings().Version or 0
	local LastCharacterUpdate = NS:GetSettings("Profile").Version or 0 -- for non-new characters, update all to current settings.
	local RunUpdate
	for i=#NewSettings, 0, -1 do
		if NewSettings[i] == false then -- global
			if i > LastGlobalUpdate then
				RunUpdate = true
				break
			end
		end
		if NewSettings[i] == true then -- per character
			if i > LastCharacterUpdate then
				RunUpdate = true
				break
			end
		end
		if i < LastGlobalUpdate and i < LastCharacterUpdate then
			break
		end
	end
	if RunUpdate then
		ToggleOptions(SetupType.NewSettings) -- New settings need to be updated
		OpenConfig = OpenConfig or NS:GetCharacterSettings().OpenConfig -- NS:GetCharacterSettings().OpenConfig can be set in FauxMazzleConfig
	end
	local ProjectId = NS:GetSettings().WOW_PROJECT_ID
	if not ProjectId or ProjectId ~= WOW_PROJECT_ID then
		ToggleOptions(0, true) -- Transitioning to new xPac. Primarily used for Classic transitions eg. vanilla-TBC/WoTLK or TBC/WoTLK etc.
	end
	NS:GetCharacterSettings().OpenConfig = nil
------------------------------------   To config. or...    ----------------------------------
	if OpenConfig then
		NS:RegisterForInit("PLAYER_ENTERING_WORLD", "OpenConfig", 4, function(self) 
			ToggleOptions(SetupType.OpenDialog + SetupType.HighlightReload) 
		end)
	end

----------------------------   Startups that require Settings    -------------------------
end

NS:RegisterForInit("PLAYER_LOGIN", "InitialiseEnvironmentWorld", 0, InitialiseEnvironmentWorld)
